package org.nakedobjects.nof.reflect.java.annotations;

import org.nakedobjects.noa.adapter.FieldRegEx;
import org.nakedobjects.noa.annotations.ActionOrderAnnotation;
import org.nakedobjects.noa.annotations.BoundedAnnotation;
import org.nakedobjects.noa.annotations.DebugAnnotation;
import org.nakedobjects.noa.annotations.DescribedAsAnnotation;
import org.nakedobjects.noa.annotations.DisabledAnnotation;
import org.nakedobjects.noa.annotations.ExecutedAnnotation;
import org.nakedobjects.noa.annotations.ExplorationAnnotation;
import org.nakedobjects.noa.annotations.FieldOrderAnnotation;
import org.nakedobjects.noa.annotations.HiddenAnnotation;
import org.nakedobjects.noa.annotations.ImmutableAnnotation;
import org.nakedobjects.noa.annotations.MaskAnnotation;
import org.nakedobjects.noa.annotations.MaxLengthAnnotation;
import org.nakedobjects.noa.annotations.MultiLineAnnotation;
import org.nakedobjects.noa.annotations.NamedAnnotation;
import org.nakedobjects.noa.annotations.NotPersistableAnnotation;
import org.nakedobjects.noa.annotations.NotPersistedAnnotation;
import org.nakedobjects.noa.annotations.OptionalAnnotation;
import org.nakedobjects.noa.annotations.PluralAnnotation;
import org.nakedobjects.noa.annotations.RegExAnnotation;
import org.nakedobjects.noa.annotations.TypeOfAnnotation;
import org.nakedobjects.noa.annotations.TypicalLengthAnnotation;
import org.nakedobjects.noa.annotations.When;
import org.nakedobjects.nof.reflect.java.reflect.JavaIntrospectorExtension;
import org.nakedobjects.nof.reflect.java.reflect.JavaRegEx;
import org.nakedobjects.nof.reflect.java.reflect.MemberHelper;

import java.lang.reflect.Method;


/**
 * 
 * REVIEW this class is intended to find indicators and data that supplement the basic reflector; not replace it.
 */

public class JavaIntrospectorAnnotationExtension implements JavaIntrospectorExtension {

    private final Class cls;

    public JavaIntrospectorAnnotationExtension(Class cls) {
        this.cls = cls;
        new JavaRegEx();
    }

    private AnnotationFactorySet annotationFactorySet = new DefaultAnnotationFactorySet();

    private AnnotationFactory getAnnotationFactory(Class annotationClass) {
        return annotationFactorySet.getFactory(annotationClass);
    }

    private Object getAnnotationOnClass(Class cls, final Class annotationClass) {
        final AnnotationFactory annotationFactory = getAnnotationFactory(annotationClass);
        Object annotationObj = annotationFactory.process(cls);
        return annotationObj;
    }

    private Object getAnnotationOnMethod(Method method, final Class annotationClass) {
        final AnnotationFactory annotationFactory = getAnnotationFactory(annotationClass);
        Object annotationObj = annotationFactory.process(method);
        return annotationObj;
    }

    private Object getAnnotationOnMethodElseReturnTypeClass(Method method, final Class annotationClass) {
        final AnnotationFactory annotationFactory = getAnnotationFactory(annotationClass);
        Object annotationObj = annotationFactory.process(method);
        if (annotationObj == null) {
            Class returnType = method.getReturnType();
            if (returnType != null) {
                annotationObj = annotationFactory.process(returnType);
            }
        }
        return annotationObj;
    }

    private Object getAnnotationOnParameterElseParamTypeClass(Method method, int paramNum, final Class annotationClass) {
        AnnotationFactory annotationFactory = getAnnotationFactory(annotationClass);
        Object annotation = annotationFactory.processParams(method, paramNum);
        if (annotation == null) {
            final Class paramType = method.getParameterTypes()[paramNum];
            annotation = annotationFactory.process(paramType);
        }
        return annotation;
    }

    // ///////////////////// @Named, @Plural ////////////////////////

    /**
     * Handle <tt>@Named</tt> for classes.
     * @return
     */
    public String getSingularName() {
        final NamedAnnotation annotation = (NamedAnnotation) getAnnotationOnClass(cls, NamedAnnotation.class);
        return annotation == null ? null : annotation.value();
    }

    /**
     * Handle <tt>@DescribedAs</tt> for classes.
     * @return
     */
    public String getDescription() {
        final DescribedAsAnnotation annotation = (DescribedAsAnnotation) getAnnotationOnClass(cls, DescribedAsAnnotation.class);
        return annotation == null ? null : annotation.value();
    }

    /**
     * Handle <tt>@Named</tt> for properties, collections and actions.
     * @return
     */
    public MemberHelper getMemberName(final Method method) {
        final NamedAnnotation annotation = (NamedAnnotation) getAnnotationOnMethod(method, NamedAnnotation.class);
        if (annotation == null) {
            return null;
        }

        return new AnnotationHelper(method, annotationFactorySet) {
            public Object execute(Object target) {
                return annotation == null ? null : annotation.value();
            }
        };
    }

    /**
     * Handle <tt>@Named</tt> for action parameters.
     */
    public MemberHelper getParameterNames(final Method method) {
        int numberOfParameters = method.getParameterTypes().length;
        final String[] names = new String[numberOfParameters];

        boolean annotated = false;
        for (int paramNum = 0; paramNum < names.length; paramNum++) {
            final NamedAnnotation annotation = (NamedAnnotation) getAnnotationOnParameterElseParamTypeClass(method, paramNum,
                    NamedAnnotation.class);
            if (annotation != null) {
                names[paramNum] = annotation.value();
                annotated = true;
            }
        }
        if (!annotated) {
            return null;
        }
        return new AnnotationHelper(method, annotationFactorySet) {
            public Object execute(Object target) {
                return names;
            }
        };
    }

    /**
     * Handle <tt>@Plural</tt> for classes.
     */
    public String getPluralName() {
        final PluralAnnotation annotation = (PluralAnnotation) getAnnotationFactory(PluralAnnotation.class).process(cls);
        return annotation == null ? null : annotation.value();
    }

    // ///////////////////// @DescribedAs ////////////////////////

    /**
     * Handle <tt>@DescribedAs</tt> for properties, collections and actions.
     */
    public MemberHelper getMemberDescription(final Method method) {
        final DescribedAsAnnotation annotation = (DescribedAsAnnotation) getAnnotationOnMethod(method,
                DescribedAsAnnotation.class);
        if (annotation == null) {
            return null;
        }
        return new AnnotationHelper(method, annotationFactorySet) {
            public Object execute(Object target) {
                return annotation == null ? null : annotation.value();
            }
        };
    }

    /**
     * Handle <tt>@DescribedAs</tt> for action parameters
     */
    public MemberHelper getParameterDescriptions(final Method method) {
        int numberOfParameters = method.getParameterTypes().length;
        final String[] descriptions = new String[numberOfParameters];

        for (int paramNum = 0; paramNum < descriptions.length; paramNum++) {
            final DescribedAsAnnotation annotation = (DescribedAsAnnotation) getAnnotationOnParameterElseParamTypeClass(method,
                    paramNum, DescribedAsAnnotation.class);
            if (annotation != null) {
                descriptions[paramNum] = annotation.value();
            }
        }
        return new AnnotationHelper(method, annotationFactorySet) {
            public Object execute(Object target) {
                return descriptions;
            }
        };
    }

    // ///////////////////// @Aggregated, @Bounded, @Cached ////////////////////////

    public boolean isAggregated() {
        /*
         * final AggregatedAnnotation annotation = (AggregatedAnnotation)
         * getAnnotationFactory(AggregatedAnnotation.class).process( cls); return annotation != null;
         */
        return false;
    }

    public boolean isBounded() {
        final BoundedAnnotation annotation = (BoundedAnnotation) getAnnotationFactory(BoundedAnnotation.class).process(cls);
        return annotation != null;
    }

    public boolean isCached() {
        /*
         * final CachedAnnotation annotation = (CachedAnnotation)
         * getAnnotationFactory(CachedAnnotation.class).process(cls); return annotation != null;
         */
        return false;
    }

    // ///////////////////// @Immutable ////////////////////////

    public boolean isImmutableOncePersisted() {
        final ImmutableAnnotation annotation = (ImmutableAnnotation) getAnnotationFactory(ImmutableAnnotation.class).process(cls);
        return annotation != null;
    }

    public boolean isAlwaysImmutable() {
        final ImmutableAnnotation annotation = (ImmutableAnnotation) getAnnotationFactory(ImmutableAnnotation.class).process(cls);
        if (annotation == null) {
            return false;
        }
        return annotation.value() == org.nakedobjects.noa.annotations.When.ALWAYS;
    }

    // ///////////////////// @FieldOrder, @ActionOrder ////////////////////////

    /**
     * Handle <tt>@FieldOrder</tt>
     * @return
     */
    public String getFieldOrder() {
        FieldOrderAnnotation annotation = (FieldOrderAnnotation) getAnnotationFactory(FieldOrderAnnotation.class).process(cls);
        if (annotation == null) {
            return null;
        }
        return annotation.value();
    }

    /**
     * Handle <tt>@ActionOrder</tt>
     */
    public String getObjectActionOrder() {
        final ActionOrderAnnotation annotation = (ActionOrderAnnotation) getAnnotationFactory(ActionOrderAnnotation.class)
                .process(cls);
        return annotation == null ? null : annotation.value();
    }

    /**
     * @deprecated
     */
    public String getClassActionOrder() {
        return null;
    }

    // ///////////////////// @Optional and @OptionalParameters ////////////////////////

    public boolean isFieldOptional(Method method) {
        AnnotationFactory annotationFactory = getAnnotationFactory(OptionalAnnotation.class);
        final OptionalAnnotation annotation = (OptionalAnnotation) annotationFactory.process(method);
        return annotation != null;
    }

    public MemberHelper getOptionalParameters(final Method method) {
        int numberOfParameters = method.getParameterTypes().length;
        final boolean[] optionals = new boolean[numberOfParameters];

        boolean annotated = false;
        for (int paramNum = 0; paramNum < optionals.length; paramNum++) {
            OptionalAnnotation annotation = (OptionalAnnotation) getAnnotationOnParameterElseParamTypeClass(method, paramNum,
                    OptionalAnnotation.class);
            if (annotation != null) {
                optionals[paramNum] = true;
                annotated = true;
            }
        }
        if (!annotated) {
            return null;
        }
        return new AnnotationHelper(method, annotationFactorySet) {
            public Object execute(Object target) {
                return optionals;
            }
        };
    }

    // ///////////////////// @NotPersisted ////////////////////////

    public boolean isFieldPersisted(Method method) {
        final NotPersistedAnnotation annotation = (NotPersistedAnnotation) getAnnotationFactory(NotPersistedAnnotation.class)
                .process(method);
        return annotation == null; // NOTE: if no annotation then the field IS persisted
    }

    public boolean isObjectTransient() {
        final NotPersistableAnnotation annotation = (NotPersistableAnnotation) getAnnotationFactory(NotPersistableAnnotation.class)
                .process(cls);
        return annotation != null && annotation.value() == NotPersistableAnnotation.By.USER_OR_PROGRAM;
    }

    public boolean isObjectNotUserPersistable() {
        final NotPersistableAnnotation annotation = (NotPersistableAnnotation) getAnnotationFactory(NotPersistableAnnotation.class)
                .process(cls);
        return annotation != null && annotation.value() == NotPersistableAnnotation.By.USER;
    }

    // ///////////////////// @Disabled ////////////////////////

    public When isActionDisabled(Method method) {
        final DisabledAnnotation annotation = (DisabledAnnotation) getAnnotationFactory(DisabledAnnotation.class).process(method);
        return annotation == null ? When.NEVER : annotation.value();
    }

    public When isFieldProtected(Method method) {
        final DisabledAnnotation annotation = (DisabledAnnotation) getAnnotationFactory(DisabledAnnotation.class).process(method);
        return annotation == null ? When.NEVER : annotation.value();
    }

    // ///////////////////// @TypeOf ////////////////////////

    /**
     * Handle <tt>@TypeOf</tt>.
     */
    public Class getCollectionType(final Method method) {
        TypeOfAnnotation annotation = (TypeOfAnnotation) getAnnotationFactory(TypeOfAnnotation.class).process(method);
        if (annotation == null) {
            return null;
        }
        return annotation.value();
    }

    // ///////////////////// @Debug, @Exploration, @Executed, @Local, @Remote ////////////////////////

    public boolean isDebug(Method method) {
        final DebugAnnotation annotation = (DebugAnnotation) getAnnotationFactory(DebugAnnotation.class).process(method);
        return annotation != null;
    }

    public boolean isExploration(Method method) {
        final ExplorationAnnotation annotation = (ExplorationAnnotation) getAnnotationFactory(ExplorationAnnotation.class)
                .process(method);
        return annotation != null;
    }

    public boolean isLocal(Method method) {
        final ExecutedAnnotation annotation = (ExecutedAnnotation) getAnnotationFactory(ExecutedAnnotation.class).process(method);
        return annotation != null && annotation.value() == ExecutedAnnotation.Where.LOCALLY;
    }

    public boolean isRemote(Method method) {
        final ExecutedAnnotation annotation = (ExecutedAnnotation) getAnnotationFactory(ExecutedAnnotation.class).process(method);
        return annotation != null && annotation.value() == ExecutedAnnotation.Where.REMOTELY;
    }

    // ///////////////////// @Hidden ////////////////////////

    public When isHidden(Method method) {
        final HiddenAnnotation annotation = (HiddenAnnotation) getAnnotationFactory(HiddenAnnotation.class).process(method);
        if (annotation == null) {
            return When.NEVER;
        } else {
            return annotation.value();
        }
    }

    // ///////////////////// @MaxLength ////////////////////////

    /**
     * Indicates that the length (maximum or typical) has not been specified.
     */
    public static final int LENGTH_NOT_SPECIFIED = 0;

    public int getMaxLength(Method method) {
        MaxLengthAnnotation annotation = (MaxLengthAnnotation) getAnnotationOnMethodElseReturnTypeClass(method,
                MaxLengthAnnotation.class);
        return annotation == null ? LENGTH_NOT_SPECIFIED : annotation.value();
    }

    public int[] getParametersMaxLengths(Method method) {
        int numberOfParameters = method.getParameterTypes().length;
        final int[] maxLengths = new int[numberOfParameters];

        for (int paramNum = 0; paramNum < maxLengths.length; paramNum++) {
            MaxLengthAnnotation annotation = (MaxLengthAnnotation) getAnnotationOnParameterElseParamTypeClass(method, paramNum,
                    MaxLengthAnnotation.class);
            if (annotation != null) {
                maxLengths[paramNum] = annotation.value();
            }
        }
        return maxLengths;
    }

    // ///////////////////// @TypicalLength ////////////////////////

    public int getTypicalLength(Method method) {
        final TypicalLengthAnnotation annotation = (TypicalLengthAnnotation) getAnnotationFactory(TypicalLengthAnnotation.class)
                .process(method);
        return annotation == null ? LENGTH_NOT_SPECIFIED : annotation.value();
    }

    public int[] getParametersTypicalLengths(Method method) {
        int numberOfParameters = method.getParameterTypes().length;
        final int[] typicalLengths = new int[numberOfParameters];

        for (int paramNum = 0; paramNum < typicalLengths.length; paramNum++) {
            TypicalLengthAnnotation annotation = (TypicalLengthAnnotation) getAnnotationOnParameterElseParamTypeClass(method,
                    paramNum, TypicalLengthAnnotation.class);
            typicalLengths[paramNum] = annotation == null ? LENGTH_NOT_SPECIFIED : annotation.value();
        }
        return typicalLengths;
    }

    // ///////////////////// @MultiLine ////////////////////////

    /**
     * Indicates that the number of lines when has not been specified.
     */
    public static final int MULTI_LINES_NUMBER_OF_LINES_DEFAULT = 1;
    public static final boolean MULTI_LINES_PREVENT_WRAPPING_DEFAULT = false;

    public int getNoLines(Method method) {
        MultiLineAnnotation annotation = (MultiLineAnnotation) getAnnotationOnMethodElseReturnTypeClass(method,
                MultiLineAnnotation.class);
        return annotation == null ? MULTI_LINES_NUMBER_OF_LINES_DEFAULT : annotation.numberOfLines();
    }

    public int[] getParametersNoLines(Method method) {
        int numberOfParameters = method.getParameterTypes().length;
        final int[] noLines = new int[numberOfParameters];

        for (int paramNum = 0; paramNum < noLines.length; paramNum++) {
            MultiLineAnnotation annotation = (MultiLineAnnotation) getAnnotationOnParameterElseParamTypeClass(method, paramNum,
                    MultiLineAnnotation.class);
            noLines[paramNum] = annotation == null ? MULTI_LINES_NUMBER_OF_LINES_DEFAULT : annotation.numberOfLines();
        }
        return noLines;
    }

    public boolean canWrap(Method method) {
        MultiLineAnnotation annotation = (MultiLineAnnotation) getAnnotationOnMethodElseReturnTypeClass(method,
                MultiLineAnnotation.class);
        return annotation == null ? MULTI_LINES_PREVENT_WRAPPING_DEFAULT : !annotation.preventWrapping();
    }

    public boolean[] getParametersCanWrap(Method method) {
        int numberOfParameters = method.getParameterTypes().length;
        final boolean[] preventWrapping = new boolean[numberOfParameters];

        for (int paramNum = 0; paramNum < preventWrapping.length; paramNum++) {
            MultiLineAnnotation annotation = (MultiLineAnnotation) getAnnotationOnParameterElseParamTypeClass(method, paramNum,
                    MultiLineAnnotation.class);
            preventWrapping[paramNum] = annotation == null ? MULTI_LINES_PREVENT_WRAPPING_DEFAULT : !annotation.preventWrapping();
        }
        return preventWrapping;
    }

    // ///////////////////// @RegEx, @Mask ////////////////////////

    public FieldRegEx getRegEx(Method method) {
        RegExAnnotation annotation = (RegExAnnotation) getAnnotationOnMethodElseReturnTypeClass(method, RegExAnnotation.class);
        return annotation == null ? null : new FieldRegEx(annotation.validation(), annotation.caseSensitive(), annotation
                .format());
    }

    public String getMask(Method method) {
        MaskAnnotation annotation = (MaskAnnotation) getAnnotationOnMethodElseReturnTypeClass(method, MaskAnnotation.class);
        return annotation == null ? null : annotation.value();
    }

    // ///////////////////// @Searchable ////////////////////////

    public static final boolean SEARCHABLE_QUERY_BY_EXAMPLE_DEFAULT = true;

    public boolean isSearchable() {
        /*
         * final SearchableAnnotation annotation = (SearchableAnnotation)
         * getAnnotationFactory(SearchableAnnotation.class).process( cls); return annotation != null;
         */
        return false;
    }

    public boolean queryByExample() {
        /*
         * final SearchableAnnotation annotation = (SearchableAnnotation)
         * getAnnotationFactory(SearchableAnnotation.class).process( cls); return annotation != null ?
         * annotation.queryByExample() : SEARCHABLE_QUERY_BY_EXAMPLE_DEFAULT;
         */
        return false;
    }

    public Class searchViaRepository() {
        /*
         * final SearchableAnnotation annotation = (SearchableAnnotation)
         * getAnnotationFactory(SearchableAnnotation.class).process( cls); return annotation != null ?
         * annotation.repository() : null;
         */
        return null;
    }

    // ///////////////////// @BusinessKey ////////////////////////

    public String getBusinessKeyName(Method method) {
        /*
         * final BusinessKeyAnnotation annotation = (BusinessKeyAnnotation)
         * getAnnotationFactory(BusinessKeyAnnotation.class) .process(cls); if (annotation == null) { return
         * null; } String name = annotation.value(); return name.equals("") ? method.getName().substring(3) :
         * name;
         */
        return null;
    }

    // ///////////////////// @Entity ////////////////////////

    public boolean isEntity() {
        // TODO add when needed as this causes a dependency to an additional library.
        // return cls.isAnnotationPresent(Entity.class);
        return false;
    }

    // ///////////////////// erm... ////////////////////////

    /**
     * Not sure what the intent is here...
     */
    public MemberHelper isUnusable(Method method) {
        return null;
    }

    /**
     * Not sure what the intent is here...
     */
    public MemberHelper getParameterDefaults(final Method method) {
        return null;
    }

    /**
     * Not sure what the intent is here...
     */
    public MemberHelper getParameterOptions(Method method) {
        return null;
    }
}

// Copyright (c) Naked Objects Group Ltd.
