package org.nakedobjects.nof.reflect.java.reflect;

import java.beans.Introspector;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.annotations.Annotation;
import org.nakedobjects.nof.reflect.java.annotations.AnnotationFactorySet;
import org.nakedobjects.nof.reflect.java.annotations.JavaIntrospectorAnnotationExtension;
import org.nakedobjects.nof.reflect.peer.MemberIdentifier;
import org.nakedobjects.nof.reflect.peer.MemberIdentifierImpl;


abstract class AbstractOneToManyStrategy implements OneToManyStrategy {
    private final static Logger LOG = Logger.getLogger(AbstractOneToManyStrategy.class);

    public void introspectAssociations(
            final JavaIntrospector introspector,
            final JavaIntrospectorAnnotationExtension extension,
            final AnnotationFactorySet annotationFactorySet,
            final Vector associations) {
        Vector v = introspector.findPrefixedMethods(JavaIntrospector.OBJECT, JavaIntrospector.GET_PREFIX, getReturnType(), 0);
        Enumeration e = v.elements();

        while (e.hasMoreElements()) {
            Method getMethod = (Method) e.nextElement();
            LOG.debug("  identified one-many association method " + getMethod);
            String capitalizedName = JavaIntrospector.javaBaseName(getMethod.getName());
            String beanName = Introspector.decapitalize(capitalizedName);

            // look for corresponding add and remove methods
            Method addMethod = introspector.findMethod(JavaIntrospector.OBJECT, "addTo" + capitalizedName, void.class, null);

            Method removeMethod = introspector.findMethod(JavaIntrospector.OBJECT, "removeFrom" + capitalizedName, void.class,
                    null);

            Class removeType = (removeMethod == null) ? null : removeMethod.getParameterTypes()[0];
            Class addType = (addMethod == null) ? null : addMethod.getParameterTypes()[0];
            Class markupType = extension.getCollectionType(getMethod);

            /*
             * The type of element can be ascertained if there is an add/associate method, otherwise it can
             * not be determined until run time.
             */
            Class elementType = getElementType(getMethod, markupType, removeType, addType);
            if (elementType == null) {
                LOG.warn("cannot determine a type for the collection " + capitalizedName + "; not added as a field");
                elementType = Object.class;
            }

            if (((addType != null) && (addType != elementType)) || ((removeType != null) && (removeType != elementType))
                    || ((markupType != null) && (markupType != elementType))) {
                LOG.error("The addTo/removeFrom methods and collection type flag in " + introspector.className() + " must "
                        + "all deal with same type of object.  There are at least two different types");
            }

            Method clearMethod = introspector.findMethod(JavaIntrospector.OBJECT, "clear" + capitalizedName, void.class, null);

            Method optionsMethod = introspector.findMethod(JavaIntrospector.OBJECT, "options" + capitalizedName, Object[].class,
                    null);

            boolean isPersisted = extension.isFieldPersisted(getMethod);

            MemberIdentifier identifier = new MemberIdentifierImpl(introspector.className(), beanName);

            FieldMethods fieldMethods = new FieldMethods(getMethod, null, addMethod, removeMethod, null, optionsMethod);

            MethodPattern methodPattern = new MethodPattern(getMethod, capitalizedName, elementType);

            Method addToValidMethod = introspector.findMethod(JavaIntrospector.OBJECT, JavaIntrospector.VALID_PREFIX + "AddTo"
                    + capitalizedName, String.class, methodPattern.getParameterTypes());
            Method removeFromValidMethod = introspector.findMethod(JavaIntrospector.OBJECT, JavaIntrospector.VALID_PREFIX
                    + "RemoveFrom" + capitalizedName, String.class, methodPattern.getParameterTypes());

            DescriptiveMethods descriptiveMethods = introspector.findDescriptiveMethods(methodPattern);
            GeneralControlMethods controlMethods = introspector.findControlMethods(JavaIntrospector.OBJECT, methodPattern,
                    addToValidMethod, removeFromValidMethod);
            FieldFlags fieldFlags = introspector.setFieldFlags(capitalizedName, getMethod, new Class[0], isPersisted, false);
            MemberSessionMethods sMethods = introspector.findSessionMethods(capitalizedName, methodPattern.getParameterTypes());

            ExtensionHolderMutableFieldPeer peer = createPeer(identifier, elementType, descriptiveMethods, fieldMethods,
                    controlMethods, fieldFlags, sMethods, clearMethod);

            final Annotation[] annotations = annotationFactorySet.process(getMethod);
            AnnotationUtil.addAnnotations(peer, annotations);

            associations.addElement(peer);
        }
    }

    protected Class getElementType(Method getMethod, Class markupType, Class removeType, Class addType) {
        Class elementType = null;
        elementType = (markupType == null) ? elementType : markupType;
        elementType = (addType == null) ? elementType : addType;
        elementType = (removeType == null) ? elementType : removeType;
        return elementType;
    }

    protected abstract ExtensionHolderMutableFieldPeer createPeer(
            final MemberIdentifier identifier,
            final Class elementType,
            final DescriptiveMethods descriptiveMethods,
            final FieldMethods fieldMethods,
            final GeneralControlMethods controlMethods,
            final FieldFlags fieldFlags,
            final MemberSessionMethods sessionMethods,
            final Method clearMethod);
}
// Copyright (c) Naked Objects Group Ltd.
