package org.nakedobjects.nof.reflect.java.reflect;



/**
 * Strategy used to determine facts about classes, such as whether an an object of a particular class can be
 * used as a field. Used for the JavaIntrospector.
 */
public interface ClassStrategy {
	/**
	 * Return the actual Class for the supplied cls.  For example, if the application uses
	 * proxies or other means to wrap a Class, then this method should return the
	 * underlying Class which should be exposed by the introspector.
	 */
	Class getClass(Class cls);
	
	/**
     * Get an array of classes which cannot be used for fields (not just values, but one-to-one associations
     * as well). Return null or an empty array if all types are valid.
     */
    Class[] getInvalidFieldTypes();

    /**
     * A list of classes which can be used for value objects. Other classes not on the getInvalidFieldTypes
     * list will be treated as associations.
     */
    Class[] getValueTypes();
    
    /**
     * Whether the supplied class is considered to have value semantics.
     * 
     * <p>
     * Clients of class strategy should call this method rather than naively iterate over
     * {@link #getValueTypes()}.
     * 
     * @param clazz
     * @return
     */
    boolean isValueType(Class clazz);
    
    /**
     * Adds a class to the list of classes (as per {@link #getValueTypes()}) that are
     * considered to have value semantics.
     * 
     * <p>
     * Added to support the <tt>@Value</tt> annotation.
     * 
     * @param clazz
     */
    public void addValueType(Class clazz);

    /**
     * Return true if the class is used by the system, and should therefore not be exposed to the user as a
     * field
     */
    boolean isSystemClass(Class cls);
}
// Copyright (c) Naked Objects Group Ltd.
