package org.nakedobjects.nof.reflect.java.reflect;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.nakedobjects.applib.ApplicationException;
import org.nakedobjects.applib.security.RoleMemento;
import org.nakedobjects.applib.security.UserMemento;
import org.nakedobjects.noa.NakedObjectApplicationException;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedReference;
import org.nakedobjects.noa.annotations.When;
import org.nakedobjects.noa.exceptions.DisabledImperativelyException;
import org.nakedobjects.noa.exceptions.DisabledNotAuthorizedException;
import org.nakedobjects.noa.reflect.Consent;
import org.nakedobjects.noa.security.Session;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.reflect.Allow;
import org.nakedobjects.nof.core.reflect.Veto;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.reflect.peer.MemberIdentifier;
import org.nakedobjects.nof.reflect.peer.ReflectionException;
import org.nakedobjects.nof.reflect.peer.ReflectiveActionException;


public abstract class JavaMember implements ExtensionHolderMutable {
    private final static Logger LOG = Logger.getLogger(JavaMember.class);
    private final MemberIdentifier identifier;
    protected final MemberHelper hiddenMethod;
    protected final MemberHelper usableMethod;
    protected final MemberHelper name;
    protected final MemberHelper description;
    protected final When isProtected;
    private final When isHidden;
    private final static Map wrapperClasses = new HashMap();

    static {
        wrapperClasses.put(boolean.class, Boolean.class);
        wrapperClasses.put(byte.class, Byte.class);
        wrapperClasses.put(short.class, Short.class);
        wrapperClasses.put(int.class, Integer.class);
        wrapperClasses.put(long.class, Long.class);
        wrapperClasses.put(float.class, Float.class);
        wrapperClasses.put(double.class, Double.class);
    }
    /**
     * ie the (static) <tt>hideXxx(UserMemento)</tt> method.
     */
    private final Method sessionHideMethod;
    /**
     * ie the (static) <tt>disableXxx(UserMemento)</tt> method.
     */
    private final Method sessionReadOnlyMethod;

    /**
     * populated during constructor.
     */
    private Vector methods = new Vector();
    /**
     * populated on first call to {@link #getMethods()} from {@link #methods}.
     */
    private Method[] methodsArray;

    protected JavaMember(
            final MemberIdentifier identifier,
            final DescriptiveMethods descriptiveMethods,
            final GeneralControlMethods controlMethods,
            final When isProtected,
            final When isHidden,
            final MemberSessionMethods sessionMethods) {
        this.identifier = identifier;
        this.hiddenMethod = remember(controlMethods.getHideMethod());
        this.usableMethod = remember(controlMethods.getusableMethod());
        this.name = remember(descriptiveMethods.getNameMethod());
        this.description = remember(descriptiveMethods.getDescriptionMethod());
        this.sessionHideMethod = remember(sessionMethods.getSessionHideMethod());
        this.sessionReadOnlyMethod = remember(sessionMethods.getSessionReadOnlyMethod());

        this.isProtected = isProtected;
        this.isHidden = isHidden;
    }

    /**
     * Adds method (represented by the supplied {@link MemberHelper}) to the array returned by
     * {@link #getMethods()}.
     */
    protected MemberHelper remember(MemberHelper memberHelper) {
        if (memberHelper instanceof MethodHelper) {
            MethodHelper methodHelper = (MethodHelper) memberHelper;
            Method method = methodHelper.getMethod();
            remember(method);
        }
        return memberHelper;
    }

    /**
     * Adds method to the array returned by {@link #getMethods()}.
     */
    protected Method remember(Method method) {
        if (method != null) {
            methods.add(method);
        }
        return method;
    }

    public void debugData(final DebugString debug) {
        debug.appendln("Identifier", identifier.toString());
        if (!(hiddenMethod instanceof NoMemberHelper)) {
            debug.appendln("Visible", hiddenMethod);
        }
        if (!(usableMethod instanceof NoMemberHelper)) {
            debug.appendln("Available", usableMethod);
        }
        if (!(name instanceof NoMemberHelper)) {
            debug.appendln("Name", name);
        }
        if (!(description instanceof NoMemberHelper)) {
            debug.appendln("Description", description);
        }
        if (sessionHideMethod != null) {
            debug.appendln("Session hide", sessionHideMethod);
        }
        if (sessionReadOnlyMethod != null) {
            debug.appendln("Authorised", sessionReadOnlyMethod);
        }
        if (isHidden != null && isHidden != When.NEVER) {
            debug.appendln("Hidden " + isHidden.getFriendlyName());
        }

    }

    protected Object domainObject(Naked inObject) {
        // NOTE if the method for this member is static then the first parameter (the instance reference) is
        // ignored
        return inObject == null ? null : inObject.getObject();
    }

    protected boolean executeBoolean(final Method method, boolean defaultFlag) {
        if (method == null) {
            return defaultFlag;
        }
        int len = method.getParameterTypes().length;
        return executeBoolean(method, defaultFlag, new Object[len]);
    }

    private boolean executeBoolean(final Method method, boolean defaultFlag, Object[] parameters) {
        try {
            return ((Boolean) method.invoke(null, parameters)).booleanValue();
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + method, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + method, ignore);
        }
        return defaultFlag;
    }

    protected boolean executeBoolean(final Method method, boolean defaultFlag, final Session session) {
        if (method == null) {
            return defaultFlag;
        }
        int len = method.getParameterTypes().length;
        Object[] parameters = new Object[len];
        parameters[0] = createUserMemento(session);
        return executeBoolean(method, defaultFlag, parameters);
    }

    protected UserMemento createUserMemento(final Session session) {
        RoleMemento[] roles = new RoleMemento[session.getRoles().length];
        for (int i = 0; i < roles.length; i++) {
            roles[i] = new RoleMemento(session.getRoles()[i]);
        }
        return new UserMemento(session.getUserName(), roles);
    }

    protected boolean executeBoolean(final MemberHelper method, final NakedReference target, boolean defaultFlag) {
        if (method == null) {
            return defaultFlag;
        }
        Boolean b = (Boolean) method.execute(domainObject(target));
        return b == null ? defaultFlag : b.booleanValue();
    }

    protected Object execute(final Method method, Object[] parameters) {
        return execute(method, null, parameters);
    }

    // TODO make all other execute methods call this
    protected Object execute(final Method method, Object target, Object[] parameters) {
        if (method == null) {
            return null;
        }
        try {
            return method.invoke(target, parameters);
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + method, e);
            return null;
        } catch (IllegalAccessException ignore) {
            throw new ReflectiveActionException(ignore);
        }
    }

    protected Consent executeConsent(final MemberHelper helper, final NakedReference target, final boolean notAuthorized) {
        if (helper == null) {
            return Allow.DEFAULT;
        }
        String vetoReason = (String) helper.execute(domainObject(target));
        return createConsent(vetoReason, notAuthorized);
    }

    protected Consent executeConsent(final Method method, final NakedReference target, Naked parameter, boolean authorizationCheck) {
        if (method == null) {
            return Allow.DEFAULT;
        }
        Object[] parameters;
        if (method.getParameterTypes().length == 1) {
            parameters = new Object[] { parameter == null ? null : parameter.getObject() };
        } else {
            // default parameters
            parameters = new Object[0];
        }
        return executeConsent(method, target, parameters, authorizationCheck);
    }

    protected Consent executeConsent(
            final Method method,
            final NakedReference target,
            Object[] parameters,
            boolean authorizationCheck) {
        if (method == null) {
            return Allow.DEFAULT;
        }
        String vetoReason = null;
        try {
            vetoReason = (String) method.invoke(target == null ? null : target.getObject(), parameters);
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + method, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + method, ignore);
        }
        return createConsent(vetoReason, authorizationCheck);
    }

    /**
     * 
     * @param vetoReason -
     *            the reason to embed as a message in the exception.
     * @param notAuthorized -
     *            if <tt>true</tt>, create a {@link DisabledNotAuthorizedException}, otherwise a
     *            {@link DisabledImperativelyException}.
     * @return
     */
    private Consent createConsent(String vetoReason, boolean authorizationCheck) {
        if (vetoReason == null) {
            return Allow.DEFAULT;
        }
        return new Veto(authorizationCheck ? new DisabledNotAuthorizedException(getName(), vetoReason)
                : new DisabledImperativelyException(getName(), vetoReason));
    }

    protected String executeString(final Method method, final Session session) {
        if (method == null) {
            return null;
        }
        try {
            Object[] parameters = new Object[method.getParameterTypes().length];
            return (String) method.invoke(null, parameters);
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + method, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + method, ignore);
        }

        return "";
    }

    public MemberIdentifier getIdentifier() {
        return identifier;
    }

    public String getHelp() {
        return "";
    }

    public String getName() {
        // example of using Dan's annotation classes - currently under review
        // NamedAnnotation an = (NamedAnnotation) getExtension(NamedAnnotation.class);
        // return an == null ? null : ("annotated " + an.value());

        return (String) name.execute(null);
    }

    public String getDescription() {
        final Object execute = description.execute(null);
        return execute == null ? "" :  (String) execute;
    }

    public Consent isUsableDeclaratively() {
        return Allow.DEFAULT;
    }

    public Consent isUsableForSession(final Session session) {
        UserMemento[] params = new UserMemento[1];
        params[0] = createUserMemento(session);
        return executeConsent(sessionReadOnlyMethod, null, params, true);
    }

    public static void invocationException(final String error, final InvocationTargetException e) {
        Throwable targetException = e.getTargetException();
        if (targetException instanceof ApplicationException) {
            // an application exception from the domain code is re-thrown as an NO exception with same
            // semantics
            throw new NakedObjectApplicationException(targetException);
        }
        LOG.error(error, targetException);
        if (targetException instanceof RuntimeException) {
            throw (RuntimeException) targetException;
        } else {
            throw new ReflectionException(targetException);
        }
    }

    public boolean isVisible(final NakedReference target) {
        if (executeBoolean(hiddenMethod, target, false)) {
            return false;
        }
        if (target == null) {
            return true;
        }
        boolean isPersistent = target.getResolveState().isPersistent();
        if (!isPersistent && isHidden == When.UNTIL_PERSISTED) {
            return false;
        }
        if (isPersistent && isHidden == When.ONCE_PERSISTED) {
            return false;
        }
        return true;
    }

    public boolean isVisibleDeclaratively() {
        if (isHidden == When.ALWAYS) {
            return false;
        }
        return true;
    }

    public boolean isVisibleForSession(Session session) {
        if (executeBoolean(sessionHideMethod, false, session)) {
            return false;
        }
        return true;
    }

    Method[] getMethods() {
        if (methodsArray == null) {
            methodsArray = convertToArray(methods);
        }
        return methodsArray;
    }

    private Method[] convertToArray(Vector methodList) {
        Method[] methodArray = new Method[methodList.size()];
        Enumeration methodEnum = methodList.elements();
        int i = 0;
        while (methodEnum.hasMoreElements()) {
            methodArray[i++] = (Method) methodEnum.nextElement();
        }
        return (Method[]) methodArray;
    }

    protected Object[] convertCharToCharacterArray(final Object originalArray) {
        final char[] original = (char[]) originalArray;
        final int len = original.length;
        final Character[] converted = new Character[len];
        for (int i = 0; i < converted.length; i++) {
            converted[i] = new Character(original[i]);
        }
        return converted;
    }

    protected Object[] convertPrimitiveToObjectArray(final Class arrayType, final Object originalArray) {
        Object[] convertedArray;
        try {
            final Class wrapperClass = (Class) wrapperClasses.get(arrayType);
            final Constructor constructor = wrapperClass.getConstructor(new Class[] { String.class });
            final int len = Array.getLength(originalArray);
            convertedArray = (Object[]) Array.newInstance(wrapperClass, len);
            for (int i = 0; i < len; i++) {
                convertedArray[i] = constructor.newInstance(new Object[] { Array.get(originalArray, i).toString() });
            }
        } catch (NoSuchMethodException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (ArrayIndexOutOfBoundsException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (IllegalArgumentException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (InstantiationException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (IllegalAccessException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (InvocationTargetException e) {
            throw new NakedObjectRuntimeException(e);
        }
        return convertedArray;
    }

    protected Object[] getObjectAsObjectArray(final Object option) {
        final Class arrayType = option.getClass().getComponentType();
        if (arrayType.isPrimitive()) {
            if (arrayType == char.class) {
                return convertCharToCharacterArray(option);
            } else {
                return convertPrimitiveToObjectArray(arrayType, option);
            }
        } else {
            return (Object[]) option;
        }
    }
    
    protected Object[] getCollectionAsObjectArray(final Object option, final NakedObjectSpecification spec) {
        final NakedCollection adapter = NakedObjectsContext.getObjectLoader().createAdapterForCollection(option, spec);
        final Enumeration e = adapter.elements();
        final Object[] optionArray = new Object[adapter.size()];
        int j = 0;
        while (e.hasMoreElements()) {
            optionArray[j++] = ((Naked) e.nextElement()).getObject();
        }

        return optionArray;
    }

    // ////////////////// Extension /////////////////////

    private Map extensionsByClass = new HashMap();

    public Object getExtension(Class cls) {
        return extensionsByClass.get(cls);
    }

    public Class[] getExtensions() {
        return (Class[]) extensionsByClass.keySet().toArray(new Class[] {});
    }

    /**
     * As per {@link ExtensionHolderMutable}.
     */
    public void addExtension(Object extension) {
        extensionsByClass.put(extension.getClass(), extension);
    }

}
// Copyright (c) Naked Objects Group Ltd.
