package org.nakedobjects.nof.reflect.java.reflect;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.FieldRegEx;
import org.nakedobjects.noa.adapter.InvalidEntryException;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.reflect.Consent;
import org.nakedobjects.noa.reflect.Formatter;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.reflect.peer.MemberIdentifier;
import org.nakedobjects.nof.reflect.peer.NullFormatter;
import org.nakedobjects.nof.reflect.peer.ReflectionException;
import org.nakedobjects.nof.reflect.peer.ValuePeer;


public class JavaValueAssociation extends JavaField implements ValuePeer {
    private final static Logger LOG = Logger.getLogger(JavaValueAssociation.class);
    private final Formatter formatter;
    private final Method validMethod;
    private Method clearMethod;
    private Method modifyMethod;
    private final int maxLength;
    private final int typicalLength;
    private final int noLines;
    private final boolean canWrap;
    private final String mask;
    private final FieldRegEx regEx;
    private final boolean canClear;

    public JavaValueAssociation(
            final MemberIdentifier identifier,
            final Class type,
            final DescriptiveMethods descriptiveMethods,
            final FieldMethods fieldMethods,
            final GeneralControlMethods controlMethods,
            final FieldFlags staticMethods,
            final MemberSessionMethods sessionMethods,
            final int maxLength,
            final int typicalLength,
            final int noLines,
            final boolean canClear,
            final boolean canWrap,
            final String mask,
            final FieldRegEx regEx) {
        super(identifier, type, descriptiveMethods, fieldMethods, controlMethods, staticMethods, sessionMethods);
        this.maxLength = maxLength;
        this.noLines = noLines;
        this.canClear = canClear;
        this.canWrap = canWrap;
        this.mask = mask;
        this.regEx = regEx;
        formatter = new NullFormatter();

        modifyMethod = fieldMethods.getAssociateMethod();
        clearMethod = fieldMethods.getDissociateMethod();
        validMethod = controlMethods.getValidMethod1();

        int t;
        if (typicalLength == 0) {
            if (maxLength > 0) {
                t = maxLength;
            } else {
                NakedValue adapter = NakedObjectsContext.getObjectLoader().createValueInstance(getSpecification());
                t = adapter.defaultTypicalLength() * getNoLines();
            }
        } else {
            t = typicalLength;
        }

        if (maxLength > 0 && typicalLength > maxLength) {
            this.typicalLength = maxLength;
        } else {
            this.typicalLength = t;
        }
    }

    public boolean canClear() {
        return canClear;
    }
    
    public boolean canWrap() {
        return canWrap;
    }

    public void clearValue(final NakedObject inObject) {
        LOG.debug("clearValue() " + inObject.getOid() + "/" + getIdentifier());
        try {
            NakedValue value = getValue(inObject);
            if (!canClear) {
                throw new InvalidEntryException("Cannot clear value");
            } else if (clearMethod != null) {
                clearMethod.invoke(inObject.getObject(), new Object[] {});
            } else if (setMethod != null) {
                if (value != null) {
                    value.clear();
                    setMethod.invoke(inObject.getObject(), new Object[] { value.getObject() });
                }
            }
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + setMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + setMethod, ignore);
        }
    }

    public void debugData(DebugString debug) {
        super.debugData(debug);
        if (validMethod != null) {
            debug.appendln("Valid method", validMethod);
        }
        if (mask != null) {
            debug.appendln("Mask", mask);
        }
        if (regEx != null) {
            debug.appendln("RegEx", regEx);
        }
    }

    public int getMaximumLength() {
        return maxLength;
    }

    public int getNoLines() {
        return noLines;
    }

    public int getTypicalLineLength() {
        return typicalLength;
    }

    public NakedValue getDefault(NakedObject fromObject) {
        if (defaultMethod == null) {
            return null;
        }

        try {
            Object obj = defaultMethod.invoke(fromObject.getObject(), new Object[0]);
            if (obj == null) {
                return null;
            } else {
                return NakedObjectsContext.getObjectLoader().createAdapterForValue(obj);
            }
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + defaultMethod, e);
            return null;
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + defaultMethod, ignore);
            throw new ReflectionException(ignore);
        }
    }

    public NakedValue getValue(final NakedObject fromObject) {
        try {
            Object obj = getMethod.invoke(fromObject.getObject(), new Object[0]);
            NakedValue adapter;
            if (obj == null) {
                NakedObjectSpecification spec = getSpecification();
                adapter = NakedObjectsContext.getObjectLoader().createValueInstance(spec);
            } else {
                adapter = NakedObjectsContext.getObjectLoader().createAdapterForValue(obj);
            }
            if (adapter == null) {
                throw new NakedObjectRuntimeException("Value field did not have a value adapter created for it: "
                        + getIdentifier() + "/" + obj);
            }
            if (mask != null) {
                adapter.setMask(mask);
            }
            if (regEx != null) {
                adapter.setRegEx(regEx);
            }
            return adapter;
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + getMethod, e);
            return null;
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + getMethod, ignore);
            throw new ReflectionException(ignore);
        }
    }

    public Formatter getFormatter() {
        return formatter;
    }

    /**
     * Set the data in an NakedObject. Passes in an existing object to for the EO to reference.
     */
    public void initValue(final NakedObject inObject, final NakedValue setValue) {
        LOG.debug("initValue() " + getIdentifier() + " " + inObject.getOid() + "/" + setValue);
        try {
            if (setMethod != null) {
                setMethod.invoke(inObject.getObject(), new Object[] { setValue == null ? null : setValue.getObject() });
            }
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + setMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + setMethod, ignore);
        }
    }

    public boolean isEmpty(final NakedObject fromObject) {
        NakedValue value = getValue(fromObject);
        return value == null || value.isEmpty();
    }

    public Consent isValueValid(final NakedObject inObject, final NakedValue value) {
        return executeConsent(validMethod, inObject, value, false);
    }

    public void setValue(final NakedObject inObject, final NakedValue setValue) {
        LOG.debug("setValue() " + inObject.getOid() + "/" + getIdentifier() + "/" + setValue.getObject());
        try {
            if (modifyMethod != null) {
                modifyMethod.invoke(inObject.getObject(), new Object[] { setValue.getObject() });
            } else if (setMethod != null) {
                setMethod.invoke(inObject.getObject(), new Object[] { setValue.getObject() });
            }
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + setMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + setMethod, ignore);
        }
    }

    public String toString() {
        String methods = (getMethod == null ? "" : "GET") + (setMethod == null ? "" : " SET");
        return "Association [name=\"" + getIdentifier() + "\", method=" + getMethod + ", methods=" + methods + ", type="
                + getSpecification() + " ]";
    }

}
// Copyright (c) Naked Objects Group Ltd.
