package org.nakedobjects.nof.reflect.java.reflect;

import java.util.Comparator;
import java.util.StringTokenizer;

import org.nakedobjects.noa.annotations.MemberOrderAnnotation;
import org.nakedobjects.noa.spec.ExtensionHolder;
import org.nakedobjects.nof.reflect.peer.MemberPeer;


/**
 * Compares by {@link MemberOrderAnnotation} obtained from each {@link MemberPeer}).
 * 
 * <p>
 * Will also compare {@link OrderSet}s; these are put after any {@link MemberPeer}s. If there is more than
 * one OrderSet then these are compared by an {@link OrderSetGroupNameComparator}.
 * 
 * <p>
 * If there is no annotation on either member, then will compare the members by name instead.
 * 
 * <p>
 * Can specify if requires that members are in the same (group) name.
 */
public class MemberOrderComparator implements Comparator {

    private boolean ensureInSameGroup;

    /**
     * 
     * @param ensureGroupIsSame
     */
    MemberOrderComparator(boolean ensureGroupIsSame) {
        this.ensureInSameGroup = ensureGroupIsSame;
    }

    private Comparator fallbackComparator = new MemberIdentifierComparator();
    private OrderSetGroupNameComparator orderSetComparator = new OrderSetGroupNameComparator(true);

    public int compare(Object o1, Object o2) {
        if (o1 instanceof MemberPeer && o2 instanceof MemberPeer) {
            return compare((MemberPeer) o1, (MemberPeer) o2);
        }
        if (o1 instanceof OrderSet && o2 instanceof OrderSet) {
            return orderSetComparator.compare(o1, o2);
        }
        if (o1 instanceof MemberPeer && o2 instanceof OrderSet) {
            return -1; // members before OrderSets.
        }
        if (o1 instanceof OrderSet && o2 instanceof MemberPeer) {
            return +1; // members before OrderSets.
        }
        throw new IllegalArgumentException("can only compare MemberPeers and OrderSets");
    }

    public int compare(MemberPeer o1, MemberPeer o2) {
        MemberOrderAnnotation m1 = getMemberOrder(o1);
        MemberOrderAnnotation m2 = getMemberOrder(o2);

        if (m1 == null && m2 == null) {
            return fallbackComparator.compare(o1, o2);
        }

        if (m1 == null && m2 != null) {
            return +1; // annotated before non-annotated
        }
        if (m1 != null && m2 == null) {
            return -1; // annotated before non-annotated
        }

        if (ensureInSameGroup && !m1.name().equals(m2.name())) {
            throw new IllegalArgumentException("Not in same group");
        }

        final StringTokenizer tokens1 = new StringTokenizer(m1.sequence(), ".", false);
        final String[] components1 = new String[tokens1.countTokens()];
        for (int i = 0; tokens1.hasMoreTokens(); i++) {
            components1[i] = tokens1.nextToken();
        }
        final StringTokenizer tokens2 = new StringTokenizer(m2.sequence(), ".", false);
        final String[] components2 = new String[tokens2.countTokens()];
        for (int i = 0; tokens2.hasMoreTokens(); i++) {
            components2[i] = tokens2.nextToken();
        }

        int length1 = components1.length;
        int length2 = components2.length;

        // shouldn't happen but just in case.
        if (length1 == 0 && length2 == 0) {
            return fallbackComparator.compare(o1, o2);
        }

        // continue to loop until we run out of components.
        int n = 0;
        while (true) {
            int length = n + 1;
            // check if run out of components in either side
            if (length1 < length && length2 >= length) {
                return -1; // o1 before o2
            }
            if (length2 < length && length1 >= length) {
                return +1; // o2 before o1
            }
            if (length1 < length && length2 < length) {
                // run out of components
                return fallbackComparator.compare(o1, o2);
            }
            // we have this component on each side
            int componentCompare = components1[n].compareTo(components2[n]);
            if (componentCompare != 0) {
                return componentCompare;
            }
            // this component is the same; lets look at the next
            n++;
        }
    }

    private MemberOrderAnnotation getMemberOrder(ExtensionHolder eh) {
        return (MemberOrderAnnotation) eh.getExtension(MemberOrderAnnotation.class);
    }

}
