package org.nakedobjects.nof.reflect.java.value;

import java.text.NumberFormat;
import java.text.ParseException;

import org.nakedobjects.applib.value.Money;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.TextEntryParseException;
import org.nakedobjects.noa.adapter.value.MoneyValue;
import org.nakedobjects.nof.core.adapter.value.AbstractValueAdapter;
import org.nakedobjects.nof.core.conf.Configuration;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.ToString;


public class MoneyAdapter extends AbstractValueAdapter implements MoneyValue {
    private static final NumberFormat DEFAULT_NUMBER_FORMAT;
    private static final NumberFormat DEFAULT_CURRENCY_FORMAT;
    private static final String DEFAULT_CURRENCY_CODE;
    private static final String LOCAL_CURRENCY_CODE;
    private Money money;

    // REVIEW how do we get the currency for the locale pre 1.4 ?
    private static String getDefaultCurrencyCode() {
        try {
            final Class cls = DEFAULT_CURRENCY_FORMAT.getClass();
            try {
                final Object currency = cls.getMethod("getCurrency", (Class[]) null).invoke(DEFAULT_CURRENCY_FORMAT,
                        (Object[]) null);
                return (String) currency.getClass().getMethod("getCurrencyCode", (Class[]) null)
                        .invoke(currency, (Object[]) null);
            } catch (final NoSuchMethodException nsm) {
                // pre java 1.4
                return "";
            }
        } catch (final Exception e) {
            throw new NakedObjectRuntimeException(e);
        }
    }

    static {
        DEFAULT_NUMBER_FORMAT = NumberFormat.getNumberInstance();
        DEFAULT_CURRENCY_FORMAT = NumberFormat.getCurrencyInstance();
        DEFAULT_NUMBER_FORMAT.setMinimumFractionDigits(DEFAULT_CURRENCY_FORMAT.getMinimumFractionDigits());
        DEFAULT_NUMBER_FORMAT.setMaximumFractionDigits(DEFAULT_CURRENCY_FORMAT.getMaximumFractionDigits());
        LOCAL_CURRENCY_CODE = getDefaultCurrencyCode();
        String property = Configuration.ROOT + "value.money.currency";
        DEFAULT_CURRENCY_CODE = NakedObjectsContext.getConfiguration().getString(property, LOCAL_CURRENCY_CODE);
    }

    public MoneyAdapter() {}

    public MoneyAdapter(Money money) {
        this();
        this.money = money;
    }

    protected String doEncode() {
        String value = String.valueOf(money.doubleValue()) + " " + money.getCurrency();
        return value;
    }

    public boolean canClear() {
        return true;
    }

    public void clear() {
        money = null;
    }

    public String getIconName() {
        return "money";
    }

    public Object getObject() {
        return money;
    }

    public Class getValueClass() {
        return Money.class;
    }

    public boolean isEmpty() {
        return money == null;
    }

    protected void doParse(String text) {
        String entry = text.trim();
        int pos = entry.lastIndexOf(' ');
        String suffix = null;
        if (pos > 1) {
            suffix = entry.substring(pos + 1).toLowerCase();
            if (suffix.length() == 3 && Character.isLowerCase(suffix.charAt(0)) && Character.isLowerCase(suffix.charAt(1))
                    && Character.isLowerCase(suffix.charAt(2))) {
                String currencyCode = suffix;
                String amount = entry.substring(0, entry.lastIndexOf(' '));
                try {
                    money = new Money(DEFAULT_NUMBER_FORMAT.parse(amount).doubleValue(), currencyCode);
                    return;
                } catch (ParseException e) {
                    throw new TextEntryParseException("Invalid money entry", e);
                }
            } else {
                throw new TextEntryParseException("Not a distinguishable money value " + text);
            }
        }

        if (money == null || money.getCurrency().equals(LOCAL_CURRENCY_CODE)) {
            try {
                double value = DEFAULT_CURRENCY_FORMAT.parse(entry).doubleValue();
                money = new Money(value, LOCAL_CURRENCY_CODE);
                return;
            } catch (ParseException e) {
                if (money != null) {
                    throw new TextEntryParseException("Not a distinguishable money value " + text, e);
                }
            }
        }

        try {
            double value = DEFAULT_NUMBER_FORMAT.parse(text).doubleValue();
            String currencyCode = money == null ? DEFAULT_CURRENCY_CODE : money.getCurrency();
            money = new Money(value, currencyCode);
            return;
        } catch (ParseException ex) {
            throw new TextEntryParseException("Not a distinguishable money value " + text, ex);
        }

        /*
         * if (suffix != null) { try { String currencyCode = money == null ? DEFAULT_CURRENCY :
         * money.getCurrency(); double value = CURRENCY_FORMAT.parse(entry).doubleValue(); money = new
         * Money(value, currencyCode); } catch (ParseException e) { throw new TextEntryParseException("Not a
         * number " + text + " for " + CURRENCY_FORMAT.getCurrency(), e); } }
         */
    }

    protected void doRestore(String data) {
        String dataString = new String(data);
        int pos = dataString.indexOf(' ');
        String amount = dataString.substring(0, pos);
        String currency = dataString.substring(pos + 1);
        money = new Money(Double.valueOf(amount).doubleValue(), currency);
    }

    public void setMask(String mask) {}

    public String titleString() {
        if (isEmpty()) {
            return "";
        }
        boolean localCurrency = LOCAL_CURRENCY_CODE.equals(money.getCurrency());
        if (localCurrency) {
            return DEFAULT_CURRENCY_FORMAT.format(money.doubleValue());
        } else {
            return DEFAULT_NUMBER_FORMAT.format(money.doubleValue()) + " " + money.getCurrency();
        }
    }

    public String toString() {
        ToString toString = new ToString(this);
        toString.append("value", money);
        toString.append("local currency", getDefaultCurrencyCode());
        return toString.toString();
    }
}

// Copyright (c) Naked Objects Group Ltd.
